// Synchronization.h
// by Eric Berdahl
// Copyright 1996,1997 Eric Berdahl
// All rights reserved
//
// For licensing instructions, see Betelgeuse.h
//
// ========================== Monitors and Conditions ==========================
//
//	MonitorLock and MonitorEntry are a pair of classes which provide mutex
//	behavior. Specifically, MonitorLock represents the lock itself, and
//	MonitorEntry represents the acquisition of that lock. MonitorEntry is
//	designed such that it can be constructed on the stack, providing easy
//	exception safety (stack unwinding will automatically release the mutex
//	if an exception blows past you). The lifetime of the lock acquisition
//	is defined by the lifetime of the MonitorEntry object. The MonitorEntry
//	constructor will block until the thread acquires the mutex.
//
//   MonitorLock gLock;  // Create some lock
//   ...
//   // This thread does not yet hold the mutex
//   {
//      MonitorEntry theLockHolder(&gLock); // Acquire the lock
//      // This thread now hold the mutex
//   }  // Exitting the scope causes the MonitorEntry to be destructed,
//      // thereby releasing the lock
//   // This thread no longer holds the mutex
//
//	MonitorCondition can be used with MonitorLock to implement a 'monitors
//	and conditions' style of synchronization. Interested readers are
//	encouraged to consult their favorite texts for examples of this style
//	of synchronization.
//
//	A MonitorCondition represents a condition variable on a given mutex,
//	upon which the condition is constructed. Calling Wait on the condition
//	causes the thread to block until the condition is met. When the
//	condition is met, you must call either Notify or Broadcast, depending
//	the desired unblocking behavior -- Notify unblocks one threads waiting
//	on the condition, while Broadcast unblocks all threads waiting on
//	the condition. Different uses of the 'monitor and condition' metaphor
//	require different uses of your condition variables.
//
//	Please note: It is a programming error to call any MonitorCondition
//	member function unless the calling thread has locked the associated
//	MonitorLock. This error is not caught by the library, but will
//	yield strange results, usually manifesting as unprotected accesses
//	both to the MonitorCondition object and the data structures protected
//	by the associated MonitorLock.
//
// ============================ Reader/Writer Locks ============================
//
//	ReadWriteLock is a semaphore which provides a solution to the multiple
//	reader/single writer problem. The number of simultaneous readers allowed
//	(must be > 0)  is passed to the constructor (unlimited in the case of
//	the default constructor). Acquiring a read lock or write lock is done
//	by constructing a ReadLockEntry or WriteLockEntry object, respectively.
//	Usage and semantics of these Entry classes is similar to that of
//	MonitorEntry, above.
//
//	The policy implemented by ReadWriteLock is a 'fair' one. Readers will
//	block if writers are waiting. Readers will preferentially unblock
//	writers. Writers will preferentially unblock readers. Blocked readers
//	may pass blocked writers.
//
// ============================== Atomic Integers ==============================
//
//	AtomicLong is a 32 bit value whose accesses, both reading and writing, are
//	guaranteed to be atomic. The values are written and read using Set and Get,
//	respectively. Also, a conditional write can be done with the SetConditional
//	member function, give you compare-and-swap functionality. Using AtomicLong
//	code can be written to do optimistic concurrency (spin locks) in the
//	following fashion:
//
//	AtomicLong gValue;
//	...
//	long newValue;
//	do
//	{
//		const long oldValue = gValue.Get();
//		newValue = ModifyValue(oldValue);
//	} while (!gValue.SetConditional(oldValue, newValue));
//
//	The Add and Subtract methods provide atomic arithmetic. The value is
//	changed atomically, and the old value is returned as the function result.

#pragma once

// Includes

#include <OS.h>

// Classes defined here

class AtomicLong;
class CheapReadWriteLock;
class MonitorCondition;
class MonitorEntry;
class MonitorLock;
class ReadLockEntry;
class ReadWriteLock;
class WriteLockEntry;

// Classes used here


//------------------------------------------------------------------------------
//	AtomicLong interface
//------------------------------------------------------------------------------


class AtomicLong
{
public:
	AtomicLong(long initialValue = 0);
	~AtomicLong();

	long	Add(long increment);
	long	Subtract(long decrement);

	long	Get() const;
	void	Set(long newValue);
	bool	SetConditional(long oldValue, long newValue);

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
				AtomicLong(const AtomicLong& copy);
	AtomicLong&	operator=(const AtomicLong& assign);

private:
	long	fValue;
};


//------------------------------------------------------------------------------
//	MonitorLock interface
//------------------------------------------------------------------------------


class MonitorLock
{
public:
					MonitorLock();
					~MonitorLock();

private:
	// MonitorEntry is a friend so it can acquire and release the lock
	// MonitorCondition is a friend so it can acquire and release the lock
	//    within the scope of its waiting and signalling.
	friend class MonitorCondition;
	friend class MonitorEntry;

		// These routines are for internal use only. See MonitorLock.cp
		// for details of their use.
	void			Acquire();
	void			Release();

	void			Reacquire();
	void			PrepareForReacquisition(long numberOfReacqs = 1);

	void			PrepareForConditions();

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
					MonitorLock(const MonitorLock& copy);
	MonitorLock&	operator=(const MonitorLock& assign);

private:
	long			fSemaphoreAtom;
	sem_id			fSemaphoreID;

	long			fNumberReentries;
	sem_id			fReentrySemaphoreID;
};


//------------------------------------------------------------------------------
//	MonitorEntry interface
//------------------------------------------------------------------------------


class MonitorEntry
{
public:
					MonitorEntry(MonitorLock* theLock);
					~MonitorEntry();

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
					MonitorEntry();
					MonitorEntry(const MonitorEntry& copy);
	MonitorEntry&	operator=(const MonitorEntry& assign);

private:
	MonitorLock* const	fLock;
};


//------------------------------------------------------------------------------
//	MonitorCondition interface
//------------------------------------------------------------------------------


class MonitorCondition
{
public:
					MonitorCondition(MonitorLock* theLock);
					~MonitorCondition();

		// Block on the condition until signalled by another client
	void			Wait();

		// Signal a single client that the condition has become true
	void			Notify();

		// Signal all clients that the condition has become true
	void			Broadcast();

		// Find out if anyone is waiting for this condition
	bool			AreWaiters() const;

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
						MonitorCondition(const MonitorCondition& copy);
	MonitorCondition&	operator=(const MonitorCondition& assign);

private:
	void			PlatformConstructor();
	void			PlatformDestructor();

	void			PlatformBlock();
	void			PlatformUnblock(long howMany);

private:
	MonitorLock* const	fLock;
	long				fNumberWaiters;
	sem_id				fConditionSemaphore;
};


//------------------------------------------------------------------------------
//	ReadWriteLock interface
//------------------------------------------------------------------------------


class ReadWriteLock
{
public:
	ReadWriteLock();
	ReadWriteLock(long maxNumberReaders);
	~ReadWriteLock();

private:
	// ReadLockEntry and WriteLockEntry are friends so they can acquire and
	// release the lock.
	friend class ReadLockEntry;
	friend class WriteLockEntry;

	void		AcquireExclusive();
	void		AcquireShared();

	void		ReleaseExclusive();
	void		ReleaseShared();

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
				ReadWriteLock(const ReadWriteLock& copy);
	ReadWriteLock&	operator=(const ReadWriteLock& assign);

private:
	enum ELockState { kFree, kReader, kWriter };
	void		AssertState(ELockState theState);

private:
	const long			fMaxNumberReaders;
	long				fNumberReaders;

	ELockState			fLockState;

	MonitorLock			fLock;
	MonitorCondition	fReaderCondition;
	MonitorCondition	fWriterCondition;
};


//------------------------------------------------------------------------------
//	WriteLockEntry interface
//------------------------------------------------------------------------------


class WriteLockEntry
{
public:
	WriteLockEntry(ReadWriteLock* aliasLock);
	WriteLockEntry(CheapReadWriteLock* aliasLock);
	~WriteLockEntry();

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
					WriteLockEntry(const WriteLockEntry& copy);
	WriteLockEntry&	operator=(const WriteLockEntry& assign);

private:
	ReadWriteLock* const		fLock;
	CheapReadWriteLock* const	fCheapLock;
};


//------------------------------------------------------------------------------
//	ReadLockEntry interface
//------------------------------------------------------------------------------


class ReadLockEntry
{
public:
	ReadLockEntry(ReadWriteLock* aliasLock);
	ReadLockEntry(CheapReadWriteLock* aliasLock);
	~ReadLockEntry();

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
					ReadLockEntry(const ReadLockEntry& copy);
	ReadLockEntry&	operator=(const ReadLockEntry& assign);

private:
	ReadWriteLock* const		fLock;
	CheapReadWriteLock* const	fCheapLock;
};


//------------------------------------------------------------------------------
//	CheapReadWriteLock interface
//------------------------------------------------------------------------------


class CheapReadWriteLock
{
public:
	CheapReadWriteLock(long maxNumberReaders = 1024);
	~CheapReadWriteLock();

private:
	// ReadLockEntry and WriteLockEntry are friends so they can acquire and
	// release the lock.
	friend class ReadLockEntry;
	friend class WriteLockEntry;

	void		AcquireExclusive();
	void		AcquireShared();

	void		ReleaseExclusive();
	void		ReleaseShared();

private:
	// These functions are not implemented. They are declared so the compiler
	// won't define them for us.
						CheapReadWriteLock(const CheapReadWriteLock& copy);
	CheapReadWriteLock&	operator=(const CheapReadWriteLock& assign);

private:
	void		PlatformConstructor();
	void		PlatformDestructor();

	void		PlatformBlockReader();
	void		PlatformUnblockReaders(long howMany);

	void		PlatformBlockWriter();
	void		PlatformUnblockWriter();

	void		PlatformAcquiredSemaphore();
	void		PlatformReleasingSemaphore();

private:
	AtomicLong	fSemaphoreData;
	const long	fMaxNumberReaders;
	sem_id		fReaderCondition;
	sem_id		fWriterCondition;
};
