// Task.h
// by Eric Berdahl
// Copyright 1996,1997 Eric Berdahl
// All rights reserved
//
// For licensing instructions, see Betelgeuse.h
//
// ThreadHandle is a simple interface for dealing with "thread-ness", often
// represented by operating systems by a thread ID. ThreadHandles for the
// currently executing thread can be constructed with the ESameThread
// construct as follows:
//
// ThreadHandle mySelf(ThreadHandle::kSameThread);
// // mySelf is now a ThreadHandle representing the current thread
//
// The default constructor creates a ThreadHandle which is guaranteed to
// never represent any thread in the system. It is thus in an invalid state,
// which can be checked with the IsValid member function.
//
// ThreadHandle(ThreadProgram* adoptProgram);
// ThreadHandle(ThreadProgram* adoptProgram, const char* threadName);
// Using either of these two constructors causes a new thread to be spawned
// in the current address space. If the constructor taking a name is used,
// the thread will be given that name (for use by whatever debugger is
// installed on the host system). The ThreadProgram object passed as an
// argument is adopted (adopt: v 1. to take responsibility for the deletion
// of another object) by the ThreadHandle thus constructed.
//
// The semantics of thread creation are very specific and useful for
// eliminating some race conditions involved in thread creation. The
// ThreadHandle constructor creates the new thread, and executes the
// ThreadProgram's Prepare member function in the context of the new thread.
// The ThreadHandle constructor will not return until the ThreadProgram's
// Prepare method has completed (either by throwing and exception or
// returning normally). If the ThreadProgram's Prepare method returns
// normally, the object's Run method will also be executed in the new
// thread.
//
// The IsValid member function can be used to test if a ThreadHandle object
// represents a thread currently executing in the system and to which the
// client has access [this last condition is important to clients on secure
// systems, e.g. Mach].
//
// ThreadHandles may be copied (via copy constructor and/or assignment
// operator) and compared to each other (via operator== and operator!=).
//
// The Kill member function will stop the represented thread from executing.
// The exact effects of this are platform dependent, but the library will stop
// the indicated thread from executing at the earliest opportunity. Kill
// should be used with care because the target thread may be stopped at an
// arbitrary point of execution, possibly with locks or other resources
// locked.
//
// WaitForDeathOf will block until the target thread stops executing.
// Calling WaitForDeathOf on an ThreadHandle whose IsValid is not a
// programming error.

#pragma once


// Includes

#include <OS.h>


// Classes defined here

class ThreadHandle;
class ThreadProgram;


// Classes used here


//------------------------------------------------------------------------------
//	ThreadHandle interface
//------------------------------------------------------------------------------


class ThreadHandle
{
public:
	enum ESameThread { kSameThread };

					ThreadHandle();
					ThreadHandle(ESameThread);
					ThreadHandle(ThreadProgram* adoptProgram);
					ThreadHandle(ThreadProgram* adoptProgram, const char* threadName);
					ThreadHandle(const ThreadHandle& copy);
					~ThreadHandle();

	ThreadHandle&	operator=(const ThreadHandle& assign);

	bool			operator==(const ThreadHandle& other) const;
	bool			operator!=(const ThreadHandle& other) const;

	bool			IsValid() const;

	void			Kill() const;
	void			WaitForDeathOf() const;

private:
	static long		ThreadNub(void* param);
	void			LaunchThread(ThreadProgram* adoptProgram, const char* threadName);

private:
	// NOTE: CodeWarrior does not support the mutable keyword, but the
	//       fThreadID field should be considered mutable.
	/* mutable */ thread_id	fThreadID;
};


//------------------------------------------------------------------------------
//	ThreadProgram interface
//------------------------------------------------------------------------------


class ThreadProgram
{
public:
					ThreadProgram();
					ThreadProgram(const ThreadProgram& copy);
	virtual			~ThreadProgram();

	ThreadProgram&	operator=(const ThreadProgram& assign);

	virtual void	Prepare();
	virtual void	Run() = 0;
};

