// CheapReadWriteLock.cp
// by Eric Berdahl
// Copyright 1997 Eric Berdahl
// All rights reserved
//
// For licensing instructions, see Betelgeuse.h
// For usage instructions, see Synchronization.h


#include "Synchronization.h"
#include "Exceptions.h"


const long	kMaxNumberWaitingReaders = (1 << 10);
const long	kMaxNumberPossibleReaders = (1 << 10);


//------------------------------------------------------------------------------
//	DataBits interface
//------------------------------------------------------------------------------


struct DataBits
{
	long	fData;

	long	GetWriters() const;
	void	DecrementWriters();

	long	GetReaders() const;
	void	IncrementReaders();
	void	DecrementReaders();
	void	SetReaders(long readers);

	long	GetWaitingReaders() const;
	void	IncrementWaitingReaders();
	void	DecrementWaitingReaders();
	void	DecrementWaitingReadersBy(long amount);
};


//------------------------------------------------------------------------------
//	DataBits implementation
//------------------------------------------------------------------------------


const long	k10BitMask = 0x000003FF;


long
DataBits::GetWriters() const
{
	return ((fData >> 20) & k10BitMask);
}


void
DataBits::DecrementWriters()
{
	fData -= (1 << 20);
}


long
DataBits::GetWaitingReaders() const
{
	return ((fData >> 10) & k10BitMask);
}


void
DataBits::IncrementWaitingReaders()
{
	fData += (1 << 10);
}


void
DataBits::DecrementWaitingReaders()
{
	fData -= (1 << 10);
}


void
DataBits::DecrementWaitingReadersBy(long amount)
{
	amount &= k10BitMask;
	fData -= (amount << 10);
}


long
DataBits::GetReaders() const
{
	return (fData & k10BitMask);
}


void
DataBits::SetReaders(long readers)
{
	long insertionMask = k10BitMask;

	fData &= (~insertionMask);
	fData |= (readers & k10BitMask);
}


void
DataBits::IncrementReaders()
{
	++fData;
}


void
DataBits::DecrementReaders()
{
	--fData;
}


//------------------------------------------------------------------------------
//	CheapReadWriteLock implementation
//------------------------------------------------------------------------------


#pragma export on


CheapReadWriteLock::CheapReadWriteLock(long maxNumberReaders) :
	fSemaphoreData(0),
	fMaxNumberReaders(maxNumberReaders)
{
	::Assertion(fMaxNumberReaders <= kMaxNumberPossibleReaders  &&
	            fMaxNumberReaders > 0,
				"CheapReadWriteLock construction error");

	PlatformConstructor();
}


CheapReadWriteLock::~CheapReadWriteLock()
{
	PlatformDestructor();
}


#pragma export off


void
CheapReadWriteLock::AcquireExclusive()
{
		// This operation is something we can do without resorting
		// to the bit fields (one of the nice things about
		// exclusive acquisition, if you ask me).
	const long oldValue = fSemaphoreData.Add(0x00100000);

	::DebugPrintf("CWRL::AcquireExclusive: flags transitioning from 0x%X to 0x%X\n",
	              oldValue, oldValue + 0x00100000);

	if (oldValue != 0)
	{
		PlatformBlockWriter();
	}

	PlatformAcquiredSemaphore();
}


void
CheapReadWriteLock::AcquireShared()
{
	bool mustBlock;
	DataBits theFlags;
	long oldValue;
	do
	{
		oldValue = theFlags.fData = fSemaphoreData.Get();

		if (theFlags.GetWriters() == 0)
		{
				// There are no writers. Check to see if we've hit
				// the maximum number of readers.
			if (theFlags.GetReaders() == fMaxNumberReaders)
			{
					// We've hit the maximum number of readers allowed.
					// We'll have to wait.
				::Assertion(theFlags.GetWaitingReaders() != kMaxNumberWaitingReaders,
				            "CheapReadWriteLock: Overran max number waiting readers");
				theFlags.IncrementWaitingReaders();
				mustBlock = true;
			}
			else
			{
				// We may acquire a read lock. Increment the number of
				// acquired readers and record that we don't
				// need to block.
				theFlags.IncrementReaders();
				mustBlock = false;
			}
		}
		else
		{
				// There are writers (either waiting or acquired),
				// so we can't add another reader. Increment the
				// number of waiting readers and record that
				// we must block.
			theFlags.IncrementWaitingReaders();
			mustBlock = true;
		}
	} while (!fSemaphoreData.SetConditional(oldValue, theFlags.fData));

	::DebugPrintf("CWRL::AcquireShared: flags transitioning from 0x%X to 0x%X (%s)\n",
	              oldValue, theFlags.fData, mustBlock ? "must block" : "success");

	if (mustBlock)
	{
		PlatformBlockReader();
	}
	else
	{
			// We only really need to do this if we were the first
			// reader to acquire the lock. But, we'll do it conservatively:
			// whenever we've done an uncontested acquisition.
		PlatformAcquiredSemaphore();
	}
}


void
CheapReadWriteLock::ReleaseExclusive()
{
	PlatformReleasingSemaphore();

	long unblockInstructions;
	DataBits theFlags;
	long oldValue;
	do
	{
		oldValue = theFlags.fData = fSemaphoreData.Get();

		theFlags.DecrementWriters();
		if (theFlags.GetWaitingReaders() == 0)
		{
			if (theFlags.GetWriters() == 0)
			{
				unblockInstructions = 0;
			}
			else
			{
				unblockInstructions = -1;
			}
		}
		else
		{
			unblockInstructions = theFlags.GetWaitingReaders();
			if (unblockInstructions > fMaxNumberReaders)
			{
				unblockInstructions = fMaxNumberReaders;
			}

			theFlags.SetReaders(unblockInstructions);
			theFlags.DecrementWaitingReadersBy(unblockInstructions);
		}
	} while (!fSemaphoreData.SetConditional(oldValue, theFlags.fData));

	::DebugPrintf("CWRL::ReleaseExclusive: flags transitioning from 0x%X to 0x%X (%ld)\n",
	              oldValue, theFlags.fData, unblockInstructions);

	if (unblockInstructions > 0)
	{
		PlatformUnblockReaders(unblockInstructions);
	}
	else
	{
		PlatformUnblockWriter();
	}
}


void
CheapReadWriteLock::ReleaseShared()
{
	PlatformReleasingSemaphore();

	long unblockInstructions;	// NOTE: No default is set here intentionally.
								//       Because of the spin-lock nature of the
								//       loop below, all code paths MUST set
								//       and appropriate value for unblockInstructions.
	DataBits theFlags;
	long oldValue;
	do
	{
		oldValue = theFlags.fData = fSemaphoreData.Get();

		theFlags.DecrementReaders();
		if (theFlags.GetWriters() != 0)
		{
			if (theFlags.GetReaders() == 0)
			{
				unblockInstructions = -1;	// unblock a writer;
			}
			else
			{
				unblockInstructions = 0;		// unblock no body
			}
		}
		else if (theFlags.GetWaitingReaders() != 0)
		{
			theFlags.DecrementWaitingReaders();
			theFlags.IncrementReaders();
			unblockInstructions = 1;		// unblock one reader
		}
		else
		{
			unblockInstructions = 0;		// unblock no body
		}
	} while (!fSemaphoreData.SetConditional(oldValue, theFlags.fData));

	::DebugPrintf("CWRL::ReleaseShared: flags transitioning from 0x%X to 0x%X (%ld)\n",
	              oldValue, theFlags.fData, unblockInstructions);

	if (unblockInstructions > 0)
	{
		PlatformUnblockReaders(1);
	}
	else if (unblockInstructions < 0)
	{
		PlatformUnblockWriter();
	}
}

