// MonitorCondition.cp
// by Eric Berdahl
// Copyright 1996,1997 Eric Berdahl
// All rights reserved
//
// For licensing instructions, see Betelgeuse.h
// For usage instructions, see Synchronization.h


#include "Exceptions.h"
#include "Synchronization.h"


#pragma export on


//------------------------------------------------------------------------------
// MonitorCondition implementation
//
// Theory of operation:
//
// All MonitorCondition synchronization operations (i.e. Wait, Notify, and
// Broadcast) must be called while the appropriate MonitorLock is held. This
// preserves the atomicity of the MonitorCondition data.
//
// Each condition holds a count of the number of people waiting. This number
// is always modified within the scope of the lock. It is incremented before
// the lock is released when blocking and is decremented by the thread doing
// the unblocking. This strategy is necessary to resolve the race condition
// of blocking and unblocking (e.g. only release the semaphore when someone
// is actually waiting).
//
// Upon returning from blocking on the condition, we need to reacquire the
// mutex. This is done by Reacquire'ing the mutex, rather than Acquire'ing it
// again. Combined with the pre-notification (provided by Notify and Broacast),
// this allows the code returning from a condition to 'remain inside the
// monitor' without allowing yet more clients inside.
//------------------------------------------------------------------------------


MonitorCondition::MonitorCondition(MonitorLock* theLock) :
	fLock(theLock),
	fNumberWaiters(0)
{
	::Assertion(fLock != NULL, "MonitorCondition: created with NULL lock");

	PlatformConstructor();

		// Tell the mutex that it will have to deal with conditions.
	fLock->PrepareForConditions();
}


MonitorCondition::~MonitorCondition()
{
	PlatformDestructor();
}


void
MonitorCondition::Wait()
{
		// It is illegal to call Wait without the associated MonitorLock
		// held. Therefore, fNumberWaiters can be accessed directly without
		// worring about atomicity.
	++fNumberWaiters;

	fLock->Release();
	PlatformBlock();
	fLock->Reacquire();
}


bool
MonitorCondition::AreWaiters() const
{
	// It is illegal to call AreWaiters without the associated MonitorLock
	// held. Therefore, fNumberWaiters can be accessed directly without
	// worring about atomicity.
	return (fNumberWaiters != 0);
}


void
MonitorCondition::Notify()
{
		// It is illegal to call Notify without the associated MonitorLock
		// held. Therefore, fNumberWaiters can be accessed directly without
		// worring about atomicity.
	if (AreWaiters())
	{
		fLock->PrepareForReacquisition();

		// Release the semaphore, thereby notifying the condition.
		--fNumberWaiters;
		PlatformUnblock(1);
	}
}


void
MonitorCondition::Broadcast()
{
		// It is illegal to call Broadcast without the associated MonitorLock
		// held. Therefore, fNumberWaiters can be accessed directly without
		// worring about atomicity.
	if (AreWaiters())
	{
		fLock->PrepareForReacquisition(fNumberWaiters);

		// Release the semaphore, thereby notifying the condition.
		const long numberOfWaitersToRelease = fNumberWaiters;
		fNumberWaiters = 0;
		PlatformUnblock(numberOfWaitersToRelease);
	}
}


#pragma export off
