// MonitorLock.Be.cp
// by Eric Berdahl
// Copyright 1996,1997 Eric Berdahl
// All rights reserved
//
// For licensing instructions, see Betelgeuse.h
// For usage instructions, see Synchronization.h


#include "Exceptions.h"
#include "Synchronization.h"
#include "SynchronizationInternal.h"

#include <Debug.h>
#include <Errors.h>
#include <OS.h>

#include <stdio.h>


//------------------------------------------------------------------------------
//	MonitorLock's Be-specific implementation
//------------------------------------------------------------------------------


#pragma export on


MonitorLock::MonitorLock()
{
		// Initialize the reentry members
	fNumberReentries = 0;
	fReentrySemaphoreID = 0;	// initialize sem_id to an invalid value

		// Initialize the semaphore atom
	fSemaphoreAtom = 0;

		// Intialize the semaphore itself
	char theName[B_OS_NAME_LENGTH];
	::sprintf(theName, "MonitorLock-%X", (long) this);

	fSemaphoreID = ::create_sem(0, theName);
	if (fSemaphoreID < B_NO_ERROR)
	{
		::FailBeError(fSemaphoreID);
	}
}


MonitorLock::~MonitorLock()
{
	::FailBeError(::delete_sem(fSemaphoreID));

		// Only delete the reentry semaphore if we've created it.
	if (fReentrySemaphoreID > 0)
	{
		::FailBeError(::delete_sem(fReentrySemaphoreID));
	}
}


#pragma export off


void
MonitorLock::PrepareForConditions()
{
		// Quick return -- we've already been initialized.
		// Must use atomic_or here because there may be another
		// thread updating fReentrySemaphoreID at the same time,
		// and we want to make sure we get a correct value.
	if (::atomic_or(&fReentrySemaphoreID, 0) != 0)
		return;

		// Attempt to create a semaphore and assign it to
		// fReentrySemaphoreID. We use an atomic 
	char theName[B_OS_NAME_LENGTH];
	::sprintf(theName, "MonitorLock-Reentry-%X", (long) this);
	const sem_id theNewSemaphore = ::create_sem(0, theName);
	if (theNewSemaphore < B_NO_ERROR)
	{
		::FailBeError(theNewSemaphore);
	}
	else if (!::CompareAndSwap(&fReentrySemaphoreID, 0, theNewSemaphore))
	{
			// We missed. Someone else has already created the
			// reentry semaphore. Delete the extra one we've made.
		::printf("semaphore = %ld\n", fReentrySemaphoreID);
		::fflush(stdout);
		::Assertion(false, "Someone missed when creating semaphores");
		::FailBeError(::delete_sem(theNewSemaphore));
	}

	::Assertion(fReentrySemaphoreID > 0, "MonitorLock::PrepareForConditions failed to create a semaphore");
}


void
MonitorLock::Acquire()
{
		// Use a simple long as a cheap lock. Attempting to
		// enter the monitor simply increments the long atomically.
		// A transition from zero to one indicates successful
		// acquisition. Other transitions require us to block.
	const long previousValue = ::atomic_add(&fSemaphoreAtom, 1);
	::Assertion(previousValue >= 0, "MonitorLock::Acquire: fSemaphoreAtom went negative!");

	if (previousValue > 0)
	{
		PRINT(("thread [%ld] blocking on MonitorLock %X\n",
		       ::find_thread(NULL), (long) this));
		::FailBeError(::acquire_sem(fSemaphoreID));
		PRINT(("thread [%ld] acquiring MonitorLock %X after contest\n",
		       ::find_thread(NULL), (long) this));
	}
	else
	{
		PRINT(("thread [%ld] acquiring MonitorLock %X without contest\n",
		       ::find_thread(NULL), (long) this));
	}
}


void
MonitorLock::Reacquire()
{
	// Reacquire is called by MonitorCondition::Wait after the condition
	// semaphore has released. It allows the thread whose condition
	// has been met to re-enter the monitor.

	::Assertion(fReentrySemaphoreID > 0, "MonitorLock::Reacquire: there isn't a semaphore");

	PRINT(("thread [%ld] attempting to reacquire MonitorLock %X\n",
	       ::find_thread(NULL), (long) this));

	::FailBeError(::acquire_sem(fReentrySemaphoreID));

	PRINT(("thread [%ld] reacquiring MonitorLock %X\n",
	       ::find_thread(NULL), (long) this));
}


void
MonitorLock::Release()
{
		// When releasing, give preference to conditions already
		// within the scope of the mutex who are returning from
		// Wait'ing.
	if (fNumberReentries > 0)
	{
		PRINT(("thread [%ld] releasing MonitorLock %X to a condition re-entry\n",
		       ::find_thread(NULL), (long) this));

		--fNumberReentries;
		::FailBeError(::release_sem(fReentrySemaphoreID));
	}
	else
	{
			// There are no re-entry attempts in the queue.
			// Release the mutex.
		const long previousValue = ::atomic_add(&fSemaphoreAtom, -1);
		::Assertion(previousValue >= 1, "MonitorLock::Release: fSemaphoreAtom went negative");
		if (previousValue > 1)
		{
			PRINT(("thread [%ld] releasing MonitorLock %X to an external waiter\n",
			       ::find_thread(NULL), (long) this));
			::FailBeError(::release_sem(fSemaphoreID));
		}
		else
		{
			PRINT(("thread [%ld] releasing MonitorLock %X uncontested\n",
			       ::find_thread(NULL), (long) this));
		}
	}
}


void
MonitorLock::PrepareForReacquisition(long numberOfReacqs)
{
	// This function is called from MonitorCondition::Broadcast
	// and from MonitorCondition::Notify before the condition
	// variable (semaphore) is signalled (released). The purpose
	// is to notify the mutex that a given number of conditions
	// can be expected to re-enter the monitor, thereby triggering
	// the alternate code path in MonitorLock::Release.
	//
	// Note: Only Conditions call this routine, and by definition,
	//       they are within the scope of the monitor at that
	//       time. Therefore, all accesses are already protected.
	::Assertion(fSemaphoreAtom > 0, "MonitorCondition trying to Notify or Broadcast when the lock is not held");

	PRINT(("thread [%ld] preparing MonitorLock %X for %ld reacquisitions\n",
	       ::find_thread(NULL), (long) this, numberOfReacqs));

	fNumberReentries += numberOfReacqs;
}

