// ReadWriteLock.cp
// by Eric Berdahl
// Copyright 1997 Eric Berdahl
// All rights reserved
//
// For licensing instructions, see Betelgeuse.h
// For usage instructions, see Synchronization.h


#include "Exceptions.h"
#include "Synchronization.h"


//------------------------------------------------------------------------------
//	ReadWriteLock implementation
//------------------------------------------------------------------------------


#pragma export on

ReadWriteLock::ReadWriteLock() :
	fMaxNumberReaders(0),
	fNumberReaders(0),
	fLockState(kFree),
	fLock(),
	fReaderCondition(&fLock),
	fWriterCondition(&fLock)
{
	// This space intentionally left blank
}


ReadWriteLock::ReadWriteLock(long maxNumberReaders) :
	fMaxNumberReaders(maxNumberReaders),
	fNumberReaders(0),
	fLockState(kFree),
	fLock(),
	fReaderCondition(&fLock),
	fWriterCondition(&fLock)
{
}


ReadWriteLock::~ReadWriteLock()
{
	// This space intentionally left blank
}

#pragma export off


void
ReadWriteLock::AcquireExclusive()
{
	MonitorEntry trojan(&fLock);

	switch (fLockState)
	{
	case kReader:
	case kWriter:
		fWriterCondition.Wait();
		AssertState(kWriter);
		break;
	
	case kFree:
		fLockState = kWriter;
		break;

	default:
		::Assertion(false, "ReadWriteLock::AcquireExclusive: bad state");
		break;
	}
}


void
ReadWriteLock::AcquireShared()
{
	MonitorEntry trojan(&fLock);

	switch (fLockState)
	{
	case kWriter:
		fReaderCondition.Wait();
		AssertState(kWriter);
		break;

	case kReader:
		if (fWriterCondition.AreWaiters())
		{
			// Don't allow more readers to acquire the lock if
			// there are waiting writers
			fReaderCondition.Wait();
			AssertState(kReader);
		}
		while (fMaxNumberReaders != 0  &&  fNumberReaders == fMaxNumberReaders)
		{
			fReaderCondition.Wait();
			AssertState(kReader);
		}
		// break intentionally omitted;

	case kFree:
		fLockState = kReader;

		++fNumberReaders;
		break;

	default:
		::Assertion(false, "ReadWriteLock::AcquireShared: bad state");
		break;
	}
}


void
ReadWriteLock::ReleaseExclusive()
{
	MonitorEntry trojan(&fLock);
	AssertState(kWriter);

	if (fReaderCondition.AreWaiters())
	{
		fLockState = kReader;
		fReaderCondition.Broadcast();
	}
	else if (fWriterCondition.AreWaiters())
	{
		fWriterCondition.Notify();
	}
	else
	{
		fLockState = kFree;
	}
}


void
ReadWriteLock::ReleaseShared()
{
	MonitorEntry trojan(&fLock);
	AssertState(kReader);

	--fNumberReaders;
	::Assertion(fMaxNumberReaders == 0  ||  fNumberReaders < fMaxNumberReaders,
	            "ReadWriteLock::ReleaseShared: bad number of readers");

	if (fWriterCondition.AreWaiters())
	{
		if (fNumberReaders == 0)
		{
			fLockState = kWriter;
			fWriterCondition.Notify();
		}
	}
	else if (fReaderCondition.AreWaiters())
	{
			// Theoretically, the only way we get here is if we've previously
			// hit the established maximum number of readers. Test for that
			// situation (i.e. that there *is* a maximum number of readers).
		::Assertion(fMaxNumberReaders > 0,
		            "ReadWriteLock::ReleaseShared: waiting readers without writers "
		            "and no max number of readers");
		fReaderCondition.Notify();
	}
	else if (fNumberReaders == 0)
	{
		fLockState = kFree;
	}
}


void
ReadWriteLock::AssertState(ELockState theState)
{
	::Assertion(fLockState == theState, "ReadWriteLock: not in expected state");
}
