// MonitorLockTestProgram.cp
// by Eric Berdahl
// Copyright 1997 Eric Berdahl
// All rights reserved

#include "MonitorLockTestProgram.h"

#include <stdio.h>
#include <stdlib.h>


//------------------------------------------------------------------------------
// MonitorLockTestProgram::Data implementation
//------------------------------------------------------------------------------


MonitorLockTestProgram::Data::Data() :
	fIncrement(1),
	fNumberIterations(1000000),
	fProtectedResource(0),
	fNumberExecutions(0),
	fLock(),
	fCondition(&fLock),
	fSuccess(true),
	fQuitting(false)
{
	// This space intentionally left blank
}


MonitorLockTestProgram::Data::~Data()
{
	// This space intentionally left blank
}


void
MonitorLockTestProgram::Data::PrintResults() const
{
	if (fSuccess  &&  fNumberExecutions != 0)
	{
		::printf("PASS\n");
	}
	else
	{
		::printf("FAIL - ");
		if (!fSuccess)
		{
			::printf("mutual exclusion failed\n");
		}
		else
		{
			::printf("complete exclusion observed\n");
		}
	}
	::fflush(stdout);
}


//------------------------------------------------------------------------------
// MonitorLockTestProgram implementation
//------------------------------------------------------------------------------


MonitorLockTestProgram::MonitorLockTestProgram() :
	ThreadProgram()
{
	// This space intentionally left blank
}


MonitorLockTestProgram::~MonitorLockTestProgram()
{
	// This space intentionally left blank
}


void
MonitorLockTestProgram::Run()
{
	const int kNumberSeconds = 10;
	::printf("MonitorLock test: running (approx %d seconds)...\n", 2*kNumberSeconds);

	TestEntry(kNumberSeconds);
	TestReentry(kNumberSeconds);

	::printf("MonitorLock test: completed.\n");
}


void
MonitorLockTestProgram::TestEntry(long numberSeconds)
{
	Data theData;

	ThreadHandle threadOne(new EntryTestProgram(&theData, true));
	ThreadHandle threadTwo(new EntryTestProgram(&theData, false));

	const long kSeconds = 1000000;
	::snooze(numberSeconds*kSeconds);

	theData.fQuitting = true;

	threadOne.WaitForDeathOf();
	threadTwo.WaitForDeathOf();

	::printf("MonitorLock test: Entry ");
	theData.PrintResults();
}


void
MonitorLockTestProgram::TestReentry(long numberSeconds)
{
	Data theData;

	ThreadHandle threadOne(new ReentryTestProgram(&theData, true));
	ThreadHandle threadTwo(new ReentryTestProgram(&theData, false));
	ThreadHandle threadThree(new ReentryTimerProgram(&theData));

	const long kSeconds = 1000000;
	::snooze(numberSeconds*kSeconds);

	theData.fQuitting = true;

	threadOne.WaitForDeathOf();
	threadTwo.WaitForDeathOf();
	threadThree.WaitForDeathOf();

	::printf("MonitorLock test: Reentry ");
	theData.PrintResults();
}


//------------------------------------------------------------------------------
// MonitorLockTestProgram::TestProgram implementation
//------------------------------------------------------------------------------


MonitorLockTestProgram::TestProgram::TestProgram(Data* theData, bool increment) :
	ThreadProgram(),
	fData(theData),
	fSignum(increment ? 1 : -1)
{
	// This space intentionally left blank
}


MonitorLockTestProgram::TestProgram::~TestProgram()
{
	// This space intentionally left blank
}


void
MonitorLockTestProgram::TestProgram::TestIteration()
{
	++fData->fNumberExecutions;

	for (long i = fData->fNumberIterations; i > 0; --i)
	{
		fData->fProtectedResource += fSignum*fData->fIncrement;
	}
	
		// Now, check the state of the protected resource.
		// If the numbers don't add up, someone else is
		// inside the monitor with us.
	if ((fData->fProtectedResource % (fData->fNumberIterations*fData->fIncrement)) != 0)
	{
			// Once we determine that the test failed, there's
			// no point in continuing.
		fData->fSuccess = false;
		fData->fQuitting = true;
	}	
}


void
MonitorLockTestProgram::TestProgram::Pause()
{
	if (!fData->fQuitting)
	{
		const long kHalfSeconds = 500000;
		const int numSeconds = ::rand() % 4;
		::snooze(numSeconds*kHalfSeconds);
	}
}


//------------------------------------------------------------------------------
// MonitorLockTestProgram::EntryTestProgram implementation
//------------------------------------------------------------------------------


MonitorLockTestProgram::EntryTestProgram::EntryTestProgram(Data* theData,
		bool increment) :
	TestProgram(theData, increment)
{
	// This space intentionally left blank
}


MonitorLockTestProgram::EntryTestProgram::~EntryTestProgram()
{
	// This space intentionally left blank
}


void
MonitorLockTestProgram::EntryTestProgram::Run()
{
	while (!fData->fQuitting)
	{
		{
				// Enter the monitor and do our really long loop
			MonitorEntry trojan(&fData->fLock);
			TestIteration();
		}

			// Just to add a little spice to the stew, we'll stagger
			// the threads, sleeping for a while. This may add contention
			// to the pool and help jog out some race conditions.
		Pause();
	}
}


//------------------------------------------------------------------------------
// MonitorLockTestProgram::ReentryTestProgram implementation
//------------------------------------------------------------------------------


MonitorLockTestProgram::ReentryTestProgram::ReentryTestProgram(Data* theData,
		bool increment) :
	TestProgram(theData, increment)
{
	// This space intentionally left blank
}


MonitorLockTestProgram::ReentryTestProgram::~ReentryTestProgram()
{
	// This space intentionally left blank
}


void
MonitorLockTestProgram::ReentryTestProgram::Run()
{
	MonitorEntry trojan(&fData->fLock);
	while (!fData->fQuitting)
	{
		fData->fCondition.Wait();
		TestIteration();

			// Just to add a little spice to the stew, we'll stagger
			// the threads, sleeping for a while. This may add contention
			// to the pool and help jog out some race conditions.
		Pause();
	}
}


//------------------------------------------------------------------------------
// MonitorLockTestProgram::ReentryTimerProgram implementation
//------------------------------------------------------------------------------


MonitorLockTestProgram::ReentryTimerProgram::ReentryTimerProgram(Data* theData) :
	TestProgram(theData, true)
{
	// This space intentionally left blank
}


MonitorLockTestProgram::ReentryTimerProgram::~ReentryTimerProgram()
{
	// This space intentionally left blank
}


void
MonitorLockTestProgram::ReentryTimerProgram::Run()
{
	const long kMilliseconds = 1000;
	const long kSeconds = kMilliseconds*1000;

	while (!fData->fQuitting)
	{
			// Enter the monitor every so often and wake
			// folks up
		::snooze(10*kMilliseconds);
		MonitorEntry trojan(&fData->fLock);
		fData->fCondition.Broadcast();
	}

		// Sleep for a couple seconds in an attempt to sew up
		// some remaining race conditions between IsQuitting
		// and Wait in ReentryTestPrograms while quitting.
	::snooze(2*kSeconds);

		// One last ping to get everyone out of their blocks
	MonitorEntry trojan(&fData->fLock);
	fData->fCondition.Broadcast();
}

