// ThreadHandle.cp
// by Eric Berdahl
// Copyright 1996,1997 Eric Berdahl
// All rights reserved
//
// For licensing instructions, see Betelgeuse.h
// For usage instructions, see Task.h


#include "Exceptions.h"
#include "Task.h"


struct ThreadLaunchInfo
{
	ThreadProgram*	fProgram;
	sem_id			fPrepareSemaphore;
};

const thread_id kInvalidThreadID = 0;


//------------------------------------------------------------------------------
//	ThreadHandle implementation
//------------------------------------------------------------------------------


#pragma export on


ThreadHandle::ThreadHandle() :
	fThreadID(kInvalidThreadID)
{
	// This space intentionally left blank
}


ThreadHandle::ThreadHandle(ESameThread) :
	fThreadID(kInvalidThreadID)
{
	fThreadID = ::find_thread(NULL);
}


ThreadHandle::ThreadHandle(ThreadProgram* adoptProgram) :
	fThreadID(kInvalidThreadID)
{
	LaunchThread(adoptProgram, B_EMPTY_STRING);
}


ThreadHandle::ThreadHandle(ThreadProgram* adoptProgram, const char* threadName) :
	fThreadID(kInvalidThreadID)
{
	LaunchThread(adoptProgram, threadName);
}



ThreadHandle::ThreadHandle(const ThreadHandle& copy) :
	fThreadID(copy.fThreadID)
{
	// This space intentionally left blank
}


ThreadHandle::~ThreadHandle()
{
	// This space intentionally left blank
}


ThreadHandle&
ThreadHandle::operator=(const ThreadHandle& assign)
{
	if (this != &assign)
	{
		fThreadID = assign.fThreadID;
	}

	return *this;
}


bool
ThreadHandle::operator==(const ThreadHandle& other) const
{
	return (fThreadID == other.fThreadID);
}


bool
ThreadHandle::operator!=(const ThreadHandle& other) const
{
	return !operator==(other);
}


bool
ThreadHandle::IsValid() const
{
		// Quick return, we are invalid
	if (fThreadID == kInvalidThreadID)
		return false;

	thread_info theThreadInfo;
	const long theError = ::get_thread_info(fThreadID, &theThreadInfo);

		// B_BAD_THREAD_ID is actually ok to return from
		// get_thread_info. It means that the thread is invalid.
	if (theError == B_BAD_THREAD_ID)
	{
			// Once we know we're invalid, remember this fact.
		const_cast<ThreadHandle*>(this)->fThreadID = kInvalidThreadID;
	}
	else
	{
			// The one error we wanted to trap was not found.
			// Do the default Fail... thing.
		::FailBeError(theError);
	}

	return (theError == B_NO_ERROR);
}


void
ThreadHandle::Kill() const
{
	const long theError = ::kill_thread(fThreadID);
	if (theError != B_BAD_THREAD_ID)
	{
		::FailBeError(theError);
	}
}


void
ThreadHandle::WaitForDeathOf() const
{
	long threadReturnValue = 0;
	const long theError = ::wait_for_thread(fThreadID, &threadReturnValue);
	if (theError != B_BAD_THREAD_ID)
	{
		::FailBeError(theError);
	}

		// We know we're invalid, remember this fact.
	const_cast<ThreadHandle*>(this)->fThreadID = kInvalidThreadID;
}


#pragma export off


long
ThreadHandle::ThreadNub(void* param)
{
	long result = 0;
	ThreadLaunchInfo& launchInfo = * reinterpret_cast<ThreadLaunchInfo*>(param);

		// Make a local copy of the ThreadProgram
		// pointer. the launchInfo object is not guaranteed
		// to live past our releasing the semaphore,
		// so anything we want to keep out of it must
		// be copied to local variables.
	ThreadProgram* const theProgram = launchInfo.fProgram;

		// Call Prepare. Record whether it succeeded or not.
		// (Success is defined as "not throwing an exception.");
	bool prepareSucceeded = true;
	try
	{
		theProgram->Prepare();
	}
	catch(...)
	{
		prepareSucceeded = false;
	}

		// Tell the launching thread that prepare is done.
	::FailBeError(::release_sem(launchInfo.fPrepareSemaphore));

		// If we successfully prepared, run the program.
	if (prepareSucceeded)
	{
		try
		{
			theProgram->Run();
		}
		catch(const BeException& theBeException)
		{
			result = theBeException.GetErrorCode();
			::fprintf(stderr, "thread [%ld] exitting with uncaught BeException(%ld)\n",
			          ::find_thread(NULL), theBeException.GetErrorCode());
			::fflush(stderr);
		}
		catch(...)
		{
			result = -1;
			::fprintf(stderr, "thread [%ld] exitting with unknown uncaught exception\n",
			          ::find_thread(NULL));
			::fflush(stderr);
		}
	}

		// We are done with the ThreadProgram, delete it.
	delete theProgram;

	return result;
}


void
ThreadHandle::LaunchThread(ThreadProgram* adoptProgram, const char* threadName)
{
		// Set up the information we pass to our thread nub
	ThreadLaunchInfo launchInfo;
	launchInfo.fProgram = adoptProgram;
	launchInfo.fPrepareSemaphore = ::create_sem(0, threadName);
	if (launchInfo.fPrepareSemaphore < B_NO_ERROR)
	{
		::FailBeError(launchInfo.fPrepareSemaphore);
	}

		// Start the new thread
	fThreadID = ::spawn_thread(ThreadNub, threadName, B_NORMAL_PRIORITY, &launchInfo);
	::FailBeError(::resume_thread(fThreadID));

		// Wait for the thread to tell us that it's preparation
		// is done
	::FailBeError(::acquire_sem(launchInfo.fPrepareSemaphore));

		// Don't forget to delete the semaphore
	::FailBeError(::delete_sem(launchInfo.fPrepareSemaphore));
}

